grammar Grammar;

program:          toplevel+ ;

toplevel:         decl | def ;

def:              'def' ID '(' formalParams? ')' '->' type '{' decl* stmt* '}' ;

formalParams:     formalParam (',' formalParam)* ;

formalParam:      ID ':' type ;

decl:             ID ':' type ';' ;

type:             'int' # IntType
                | 'bool' # BoolType
                | 'string' # StringType
                | '(' paramTypes? ')' '->' returnType=type # FunType
                ;

paramTypes:       type (',' type)* ;

stmt:             name '=' expr ';' # Assignment
                | 'while' '(' expr ')' stmt # While
                // resolve dangling-else by attempting to match ifthenelse first first
                | 'if'    '(' expr ')' stmt 'else' stmt # IfThenElse
                | 'if'    '(' expr ')' stmt # IfThen
                | 'return' expr ';' # Return
                | expr ';' # ExprStmt
                | ';' # Empty
                | '{' stmt* '}' # Compound
                ;

expr:             name '(' actualParams? ')' # Call
                | '-' expr # Negate
                | '!' expr # Not
                | expr op=('*'|'/') expr # MultDiv
                | expr op=('+'|'-') expr # AddSub
                | expr op=('<' | '>' | '<=' | '>=' ) expr # Relational
                | expr op=('=='|'!=') expr # EqNeq
                | expr op=('&&'|'||') expr # AndOr
                | name # Var
                | NUM # Num
                | StringLiteral # StringLiteral
                | '(' expr ')' # Parens
                ;

name:             ID # Simple
                ;

actualParams:     expr (',' expr)* ;


// lexical spec
ID: [a-zA-Z] ([a-zA-Z] | [0-9])* ;
NUM: [0-9]+ ;
MULT: '*' ;
DIV: '/' ;
PLUS: '+' ;
MINUS: '-' ;
AND: '&&' ;
OR: '||' ;
EQ: '==' ;
NEQ: '!=' ;
LT: '<' ;
GT: '>' ;
LE: '<=' ;
GE: '>=' ;
WS:   [ \t\n\r]+ -> skip ;
SL_COMMENT:   '//' .*? '\n' -> skip ;

// adapted from https://github.com/antlr/grammars-v4/blob/master/c/C.g4
StringLiteral
    :   '"' SCharSequence? '"'
    ;
fragment
SCharSequence
    :   SChar+
    ;
fragment
SChar
    :   ~["\\\r\n]
    |   EscapeSequence
    |   '\\\n'   // Added line
    |   '\\\r\n' // Added line
    ;
fragment
EscapeSequence
    :   SimpleEscapeSequence
    ;
fragment
SimpleEscapeSequence
    :   '\\' ['"?abfnrtv\\]
    ;
