import java.util.Map;
import java.util.HashMap;

public class Scope {
  /** The parent scope.  null is the current scope is the global scope. */
  Scope parent;

  /** The symbols defined in this scope. */
  Map<String, Type> symbols = new HashMap<String, Type>();

  /** The local scopes defined in this scope. */
  Map<String, Scope> scopes = new HashMap<String, Scope>();

  /** Create a new global scope. */
  public Scope() {
    this.parent = null;
  }

  /** Create a new nested scope where parent is the non-null parent scope. */
  public Scope(Scope parent) {
    assert null != parent;
    this.parent = parent;
  }

  /** Return true if this is the global scope. */
  public boolean isGlobal() {
    return null == this.parent;
  }

  /** Return the parent scope as long as the scope is not global. */
  public Scope getParent() {
    assert ! isGlobal();
    return this.parent;
  }

  /** Check whether the symbol is defined in the local scope. */
  public boolean hasSymbol(String sym) {
    return this.symbols.containsKey(sym);
  }

  /** Add the symbol to the local scope. */
  public void addSymbol(String sym, Type type) {
    assert ! hasSymbol(sym);
    this.symbols.put(sym, type);
  }

  /** Get the symbol's type from the local scope. */
  public Type getSymbol(String sym) {
    assert hasSymbol(sym);
    return this.symbols.get(sym);
  }

  /** Check whether the symbol is defined in the local or any parent scope. */
  public boolean hasSymbolAnyScope(String sym) {
    if (hasSymbol(sym)) {
      return true;
    } else {
      if (! isGlobal()) {
        return getParent().hasSymbolAnyScope(sym);
      } else {
        return false;
      }
    }
  }

  /** Get the symbol's type, recursively searching parent scopes. */
  public Type getTypeAnyScope(String sym) {
    assert hasSymbolAnyScope(sym);

    if (hasSymbol(sym)) {
      return getSymbol(sym);
    } else {
      if (! isGlobal()) {
        return getParent().getTypeAnyScope(sym);
      } else {
        throw new AssertionError("FATAL: getTypeAnyScope 2");
      }
    }
  }

  /** Add the nested function's scope. */
  public void addScope(String sym, Scope scope) {
    assert ! hasScope(sym);
    this.scopes.put(sym, scope);
  }

  /** Check whether the function has a nested scope. */
  public boolean hasScope(String sym) {
    return this.scopes.containsKey(sym);
  }

  /** Get the nested function's scope. */
  public Scope getScope(String sym) {
    assert hasScope(sym);
    return this.scopes.get(sym);
  }

  /** Check whether the function has a nested scope. */
  public boolean hasScopeAny(String sym) {
    if (hasScope(sym)) {
      return true;
    } else {
      if (! isGlobal()) {
        return getParent().hasScopeAny(sym);
      } else {
        return false;
      }
    }
  }

  /** Get the nested function's scope. */
  public Scope getScopeAny(String sym) {
    assert hasScopeAny(sym);

    if (hasScope(sym)) {
      return getScope(sym);
    } else {
      if (! isGlobal()) {
        return getParent().getScopeAny(sym);
      } else {
        throw new AssertionError("FATAL: getTypeAnyScope 2");
      }
    }
  }
}
