import java.util.List;

import org.antlr.v4.runtime.ParserRuleContext;
import org.antlr.v4.runtime.Token;

public abstract class Type {
  enum Kind {
    PRIMITIVE,
    FUNCTION,
    CLASS
  }  

  /** Construction not permitted.  Instead use built-in and function types. */
  private Type() {
  }

  /** The kind of type. */
  public abstract Kind kind();

  /** Check for type equality. */
  public abstract boolean equals(Type type);

  /** Generate a C anonymous type string. */
  public abstract String getAnonymousType();
  
  /** Generate a C reference type string. */
  public abstract String getReferenceType(String name);
  
  /** Generate a C declaration type string. */
  public abstract String getDeclarationType(String name);
  
  /** Generate a C struct field type string. */
  public abstract String getFieldType(String name, String className);
  
  public abstract String toString();

  /** A singleton for the void type. */
  public static final Type VOID = new Type() {
      public Kind kind() { return Kind.PRIMITIVE; }
      public boolean equals(Type type) { return this == type; }
      public String getAnonymousType() { return "void"; }
      public String getReferenceType(String name) { return String.format("%s %s", getAnonymousType(), name); }
      public String getDeclarationType(String name) { return getReferenceType(name); }
      public String getFieldType(String name, String className) { return getReferenceType(name); }
      public String toString() { return "void"; }
    };

  /** A singleton for the integer type. */
  public static final Type INT = new Type() {
      public Kind kind() { return Kind.PRIMITIVE; }
      public boolean equals(Type type) { return this == type; }
      public String getAnonymousType() { return "int"; }
      public String getReferenceType(String name) { return String.format("%s %s", getAnonymousType(), name); }
      public String getDeclarationType(String name) { return getReferenceType(name); }
      public String getFieldType(String name, String className) { return getReferenceType(name); }
      public String toString() { return "int"; }
    };
    
  /** A singleton for the boolean type. */
  public static final Type BOOL = new Type() {
      public Kind kind() { return Kind.PRIMITIVE; }
      public boolean equals(Type type) { return this == type; }
      public String getAnonymousType() { return "bool"; }
      public String getReferenceType(String name) { return String.format("%s %s", getAnonymousType(), name); }
      public String getDeclarationType(String name) { return getReferenceType(name); }
      public String getFieldType(String name, String className) { return getReferenceType(name); }
      public String toString() { return "bool"; }
    };
    
  /** A singleton for the string type. */
  public static final Type STRING = new Type() {
      public Kind kind() { return Kind.PRIMITIVE; }
      public boolean equals(Type type) { return this == type; }
      public String getAnonymousType() { return "char *"; }
      public String getReferenceType(String name) { return String.format("%s %s", getAnonymousType(), name); }
      public String getDeclarationType(String name) { return getReferenceType(name); }
      public String getFieldType(String name, String className) { return getReferenceType(name); }
      public String toString() { return "string"; }
    };

  /** A class for function type. */
  public static class FunctionType extends Type {
    public final List<Type> paramTypes;
    public final Type returnType;

    public FunctionType(List<Type> paramTypes, Type returnType) {
      assert null != paramTypes;
      assert null != returnType;

      this.paramTypes = paramTypes;
      this.returnType = returnType;
    }
    
    public Kind kind() { return Kind.FUNCTION; }
    
    public boolean equals(Type type) {
      Type type1 = this;
      Type type2 = type;
      if (type1.kind() == Kind.FUNCTION && type2.kind() == Kind.FUNCTION) {
        FunctionType funType1 = (FunctionType) type1;
        FunctionType funType2 = (FunctionType) type2;
        if (funType1.paramTypes.size() == funType2.paramTypes.size()) {
          for (int idx = 0; idx < funType1.paramTypes.size(); idx++) {
            if (funType1.paramTypes.get(idx).equals(funType2.paramTypes.get(idx))) {
              // types match, continue checking
            } else {
              return false;
            }
          }
          return funType1.returnType.equals(funType2.returnType);
        } else {
          return false;
        }
      } else {
        return false;
      }
    }

    public String getAnonymousType() {
      return getReferenceType("");
    }

    public String getReferenceType(String name) {
      return getDeclarationType(String.format("(*%s)", name));
    }

    public String getDeclarationType(String name) {
      StringBuilder sb = new StringBuilder();
      sb.append(this.returnType.getAnonymousType());
      sb.append(" ");
      sb.append(name);
      sb.append(" ");
      sb.append("(");
      String delim = "";
      for (Type paramType : this.paramTypes) {
        sb.append(delim);
        sb.append(paramType.getAnonymousType());
        delim = ", ";
      }
      sb.append(")");
      return sb.toString();
    }

    public String getFieldType(String name, String className) {
      StringBuilder sb = new StringBuilder();
      sb.append(this.returnType.getAnonymousType());
      sb.append(" ");
      sb.append(String.format("(* %s)", name));
      sb.append(" ");
      sb.append("(");
      // add the extra parameter for "this"
      sb.append("struct");
      sb.append(" ");
      sb.append(className);
      sb.append(" ");
      sb.append("*");
      String delim = ", ";
      for (Type paramType : this.paramTypes) {
        sb.append(delim);
        sb.append(paramType.getAnonymousType());
        delim = ", ";
      }
      sb.append(")");
      return sb.toString();
    }

    public String toString() {
      StringBuilder sb = new StringBuilder();

      sb.append("(");
      String delim = "";
      for (Type type : this.paramTypes) {
        sb.append(delim);
        sb.append(type);
        delim = ",";
      }
      sb.append(")");
      sb.append(" -> ");
      sb.append(this.returnType);
      return sb.toString();
    }
  }

  /** A class for function type. */
  public static class ClassType extends Type {
    public final String name;

    public ClassType(String name) {
      this.name = name;
    }
    
    public Kind kind() { return Kind.CLASS; }

    public boolean equals(Type type) {
      Type type1 = this;
      Type type2 = type;
      if (type1.kind() == Kind.CLASS && type2.kind() == Kind.CLASS) {
        ClassType funType1 = (ClassType) type1;
        ClassType funType2 = (ClassType) type2;
        return funType1.name.equals(funType2.name);
      } else {
        return false;
      }
    }

    public String getAnonymousType() {
      StringBuilder sb = new StringBuilder();
      sb.append("struct");
      sb.append(" ");
      sb.append(this.name);
      sb.append(" ");
      sb.append("*");
      return sb.toString();
    }

    public String getReferenceType(String name) {
      return String.format("%s %s", getAnonymousType(), name);
    }

    public String getDeclarationType(String name) {
      return getReferenceType(name);
    }

    public String getFieldType(String name, String className) {
      return getReferenceType(name);
    }

    public String toString() {
      StringBuilder sb = new StringBuilder();

      sb.append("class");
      sb.append(" ");
      sb.append(this.name);
      return sb.toString();
    }
  }

  /** Reports a type error. */
  public static void typeError(ParserRuleContext ctx, String msg) {
    Token firstToken = ctx.getStart();
    String file = firstToken.getTokenSource().getSourceName();
    int line = firstToken.getLine();
    int col = firstToken.getCharPositionInLine() + 1;
    System.out.println(String.format("%s:%d:%d type error: %s", file, line, col, msg));
    System.exit(2);
  }
}

